"""Represent a light."""
from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from pydantic import Field

from ..color import supported_features
from ..const import (
    ATTR_DEVICE_STATE,
    ATTR_LIGHT_COLOR_HEX,
    ATTR_LIGHT_COLOR_HUE,
    ATTR_LIGHT_COLOR_SATURATION,
    ATTR_LIGHT_COLOR_X,
    ATTR_LIGHT_COLOR_Y,
    ATTR_LIGHT_DIMMER,
    ATTR_LIGHT_MIREDS,
    SUPPORT_BRIGHTNESS,
    SUPPORT_COLOR_TEMP,
    SUPPORT_HEX_COLOR,
    SUPPORT_XY_COLOR,
)
from ..resource import BaseResponse

if TYPE_CHECKING:
    # avoid cyclic import at runtime.
    from . import Device


class LightResponse(BaseResponse):
    """Represent API response for a blind."""

    color_mireds: Optional[int] = Field(alias=ATTR_LIGHT_MIREDS)
    color_hex: Optional[str] = Field(alias=ATTR_LIGHT_COLOR_HEX)
    color_xy_x: Optional[int] = Field(alias=ATTR_LIGHT_COLOR_X)
    color_xy_y: Optional[int] = Field(alias=ATTR_LIGHT_COLOR_Y)
    color_hue: Optional[int] = Field(alias=ATTR_LIGHT_COLOR_HUE)
    color_saturation: Optional[int] = Field(alias=ATTR_LIGHT_COLOR_SATURATION)
    dimmer: int = Field(alias=ATTR_LIGHT_DIMMER)
    state: int = Field(alias=ATTR_DEVICE_STATE)


class Light:
    """Represent a light.

    https://github.com/IPSO-Alliance/pub/blob/master/docs/IPSO-Smart-Objects.
    pdf
    """

    def __init__(self, device: Device, index: int) -> None:
        """Create object of class."""
        self.device = device
        self.index = index

    @property
    def supported_features(self) -> int:
        """Return supported features."""
        return supported_features(self.raw)

    @property
    def state(self) -> bool:
        """Return device state."""
        return self.raw.state == 1

    @property
    def dimmer(self) -> int | None:
        """Return dimmer if present."""
        if self.supported_features & SUPPORT_BRIGHTNESS:
            return self.raw.dimmer
        return None

    @property
    def color_temp(self) -> int | None:
        """Return color temperature."""
        if self.supported_features & SUPPORT_COLOR_TEMP and self.raw.color_mireds:
            return self.raw.color_mireds
        return None

    @property
    def hex_color(self) -> str | None:
        """Return hex color."""
        if self.supported_features & SUPPORT_HEX_COLOR:
            return self.raw.color_hex
        return None

    @property
    def xy_color(self) -> tuple[int, int] | None:
        """Return xy color."""
        if (
            self.supported_features & SUPPORT_XY_COLOR
            and self.raw.color_xy_x is not None
            and self.raw.color_xy_y is not None
        ):
            return (self.raw.color_xy_x, self.raw.color_xy_y)
        return None

    @property
    def hsb_xy_color(
        self,
    ) -> tuple[int, int, int, int, int] | None:
        """Return hsb xy color."""
        if (
            self.raw.color_hue is not None
            and self.raw.color_saturation is not None
            and self.raw.dimmer is not None
            and self.raw.color_xy_x is not None
            and self.raw.color_xy_y is not None
        ):
            return (
                self.raw.color_hue,
                self.raw.color_saturation,
                self.raw.dimmer,
                self.raw.color_xy_x,
                self.raw.color_xy_y,
            )

        return None

    @property
    def raw(self) -> LightResponse:
        """Return raw data that it represents."""
        light_control_response = self.device.raw.light_control
        assert light_control_response is not None
        return light_control_response[self.index]

    def __repr__(self) -> str:
        """Return representation of class object."""
        state = "on" if self.state else "off"
        return (
            f"<Light #{self.index} - name: {self.device.name}, state: {state}, "
            f"dimmer: {self.dimmer}, "
            f"hex_color: {self.hex_color}, "
            f"xy_color: {self.xy_color}, "
            f"hsb_xy_color: {self.hsb_xy_color}, "
            f"supported features: {self.supported_features} "
            ">"
        )



"""Class to control the signal repeater."""
from __future__ import annotations

from .base_controller import BaseController
from .signal_repeater import SignalRepeater, SignalRepeaterResponse


class SignalRepeaterControl(BaseController):
    """Class to control the signal repeaters."""

    @property
    def raw(self) -> list[SignalRepeaterResponse]:
        """Return raw data that it represents."""
        signal_repeater_control_response = self._device.raw.signal_repeater_control
        assert signal_repeater_control_response is not None
        return signal_repeater_control_response

    @property
    def signal_repeaters(self) -> list[SignalRepeater]:
        """Return signal repeater objects of the signal repeater control."""
        return [SignalRepeater(self._device, i) for i in range(len(self.raw))]



"""Represent the gateway."""
from __future__ import annotations

from datetime import datetime
from typing import Any, Optional

from pydantic import BaseModel, Field

from .command import Command
from .const import (
    ATTR_ALEXA_PAIR_STATUS,
    ATTR_AUTH,
    ATTR_CERTIFICATE_PROV,
    ATTR_COMMISSIONING_MODE,
    ATTR_CURRENT_TIME_ISO8601,
    ATTR_CURRENT_TIME_UNIX,
    ATTR_FIRMWARE_VERSION,
    ATTR_FIRST_SETUP,
    ATTR_GATEWAY_FACTORY_DEFAULTS,
    ATTR_GATEWAY_ID,
    ATTR_GATEWAY_INFO,
    ATTR_GATEWAY_REBOOT,
    ATTR_GATEWAY_TIME_SOURCE,
    ATTR_GATEWAY_UPDATE_PROGRESS,
    ATTR_GOOGLE_HOME_PAIR_STATUS,
    ATTR_HOMEKIT_ID,
    ATTR_IDENTITY,
    ATTR_NTP,
    ATTR_OTA_TYPE,
    ATTR_OTA_UPDATE_STATE,
    ATTR_PSK,
    ROOT_DEVICES,
    ROOT_GATEWAY,
    ROOT_GROUPS,
    ROOT_MOODS,
    ROOT_SMART_TASKS,
)
from .device import Device
from .group import Group
from .mood import Mood
from .resource import TypeRaw
from .smart_task import SmartTask


class GatewayInfoResponse(BaseModel):
    """Represent API response for the gateway."""

    certificate_provisioned: int = Field(alias=ATTR_CERTIFICATE_PROV)
    current_time: Optional[int] = Field(alias=ATTR_CURRENT_TIME_UNIX)
    current_time_iso8601: str = Field(alias=ATTR_CURRENT_TIME_ISO8601)
    commissioning_mode: int = Field(alias=ATTR_COMMISSIONING_MODE)
    firmware_version: str = Field(alias=ATTR_FIRMWARE_VERSION)
    first_setup: Optional[int] = Field(alias=ATTR_FIRST_SETUP)
    homekit_id: str = Field(alias=ATTR_HOMEKIT_ID)
    id: str = Field(alias=ATTR_GATEWAY_ID)
    ntp_server: str = Field(alias=ATTR_NTP)
    ota_type: int = Field(alias=ATTR_OTA_TYPE)
    ota_update_state: int = Field(ATTR_OTA_UPDATE_STATE)
    pair_status_alexa: int = Field(alias=ATTR_ALEXA_PAIR_STATUS)
    pair_status_google_home: int = Field(ATTR_GOOGLE_HOME_PAIR_STATUS)
    time_source: int = Field(alias=ATTR_GATEWAY_TIME_SOURCE)
    update_progress: int = Field(alias=ATTR_GATEWAY_UPDATE_PROGRESS)


class Gateway:
    """This class connects to the IKEA Tradfri Gateway."""

    @classmethod
    def generate_psk(cls, identity: str) -> Command[str]:
        """Generate the PRE_SHARED_KEY from the gateway.

        Returns a Command.
        """

        def process_result(result: dict[str, str]) -> str:
            return result[ATTR_PSK]

        return Command(
            "post",
            [ROOT_GATEWAY, ATTR_AUTH],
            {ATTR_IDENTITY: identity},
            process_result=process_result,
        )

    @classmethod
    def get_endpoints(cls) -> Command[list[str]]:
        """
        Return all available endpoints on the gateway.

        The response from the gateway looks like this:
        <//15006>;ct=0;obs,
        <//15001>;ct=0;obs,
        <//15004>;ct=0;obs,
        <//15004/add>;ct=0,
        <//15004/remove>;ct=0,
        <//15010>;ct=0;obs,
        <//15005>;ct=0;obs,
        <//15011/15012>;ct=0;obs,
        <//15011/9034>;ct=0,
        <//15011/9030>;ct=0,
        <//15011/9031>;ct=0,
        <//15011/9094>;ct=0;obs,
        <//15011/9095>;ct=0;obs,
        <//15011/9104>;ct=0;obs,
        <//15004/131073>;ct=0;obs,
        <//15001/65536>;ct=0;obs,
        <//15005/131073/196608>;ct=0;obs,
        <//15011/9063>;ct=0

        Returns a Command.
        """

        def process_result(result: str) -> list[str]:
            return [line.split(";")[0][2:-1] for line in result.split(",")]

        return Command(
            "get",
            [".well-known", "core"],
            parse_json=False,
            process_result=process_result,
        )

    def get_devices(self) -> Command[list[Command[Device]]]:
        """
        Return the devices linked to the gateway.

        Returns a Command.
        """

        def process_result(result: list[str]) -> list[Command[Device]]:
            return [self.get_device(dev) for dev in result]

        return Command("get", [ROOT_DEVICES], process_result=process_result)

    @classmethod
    def get_device(cls, device_id: str) -> Command[Device]:
        """
        Return specified device.

        Returns a Command.
        """

        def process_result(result: TypeRaw) -> Device:
            return Device(result)

        return Command("get", [ROOT_DEVICES, device_id], process_result=process_result)

    def get_groups(self) -> Command[list[Command[Group]]]:
        """
        Return the groups linked to the gateway.

        Returns a Command.
        """

        def process_result(result: list[str]) -> list[Command[Group]]:
            return [self.get_group(group) for group in result]

        return Command("get", [ROOT_GROUPS], process_result=process_result)

    def get_group(self, group_id: str) -> Command[Group]:
        """
        Return specified group.

        Returns a Command.
        """

        def process_result(result: TypeRaw) -> Group:
            return Group(self, result)

        return Command("get", [ROOT_GROUPS, group_id], process_result=process_result)

    @classmethod
    def add_group_member(cls, values: dict[str, Any]) -> Command[None]:
        """Add a device to a group."""

        return Command("put", [ROOT_GROUPS, "add"], values)

    @classmethod
    def remove_group_member(cls, values: dict[str, Any]) -> Command[None]:
        """Remove a device from a group."""

        return Command("put", [ROOT_GROUPS, "remove"], values)

    @classmethod
    def get_gateway_info(cls) -> Command[GatewayInfo]:
        """
        Return the gateway info.

        Returns a Command.
        """

        def process_result(result: TypeRaw) -> GatewayInfo:
            return GatewayInfo(result)

        return Command(
            "get", [ROOT_GATEWAY, ATTR_GATEWAY_INFO], process_result=process_result
        )

    def get_moods(self, group_id: int) -> Command[list[Command[Mood]]]:
        """
        Return moods available in given group.

        Returns a Command.
        """

        def process_result(result: list[str]) -> list[Command[Mood]]:
            return [self.get_mood(mood, mood_parent=group_id) for mood in result]

        return Command(
            "get", [ROOT_MOODS, str(group_id)], process_result=process_result
        )

    @classmethod
    def get_mood(cls, mood_id: str, *, mood_parent: int) -> Command[Mood]:
        """
        Return a mood.

        Returns a Command.
        """

        def process_result(result: TypeRaw) -> Mood:
            return Mood(result, mood_parent)

        return Command(
            "get",
            [ROOT_MOODS, str(mood_parent), mood_id],
            mood_parent,
            process_result=process_result,
        )

    def get_smart_tasks(self) -> Command[list[Command[SmartTask]]]:
        """
        Return the transitions linked to the gateway.

        Returns a Command.
        """

        def process_result(result: list[str]) -> list[Command[SmartTask]]:
            return [self.get_smart_task(task) for task in result]

        return Command("get", [ROOT_SMART_TASKS], process_result=process_result)

    def get_smart_task(self, task_id: str) -> Command[SmartTask]:
        """
        Return specified transition.

        Returns a Command.
        """

        def process_result(result: TypeRaw) -> SmartTask:
            return SmartTask(self, result)

        return Command(
            "get", [ROOT_SMART_TASKS, task_id], process_result=process_result
        )

    @classmethod
    def reboot(cls) -> Command[None]:
        """Reboot the Gateway.

        Returns a Command.
        """

        return Command("post", [ROOT_GATEWAY, ATTR_GATEWAY_REBOOT])

    @classmethod
    def set_commissioning_timeout(cls, timeout: int) -> Command[None]:
        """Put the gateway in pairing state.

        The pairing state is when the gateway accepts pairings from
        switches, dimmers and motion sensors for up to timeout seconds.
        Returns a Command.
        """

        return Command(
            "put", [ROOT_GATEWAY, ATTR_GATEWAY_INFO], {ATTR_COMMISSIONING_MODE: timeout}
        )

    @classmethod
    def factory_reset(cls) -> Command[None]:
        """Reset Gateway to factory defaults.

        WARNING: All data in Gateway is lost (pairing, groups, etc)
        Returns a Command.
        """

        return Command("post", [ROOT_GATEWAY, ATTR_GATEWAY_FACTORY_DEFAULTS])


class GatewayInfo:
    """This class contains Gateway information."""

    raw: GatewayInfoResponse

    def __init__(self, raw: TypeRaw) -> None:
        """Create object of class."""
        self.raw = GatewayInfoResponse(**raw)

    @property
    def certificate_provisioned(self) -> int:
        """Return provisioning status of certificate."""
        return self.raw.certificate_provisioned

    @property
    def current_time(self) -> datetime | None:
        """Return current time (normal timestamp)."""
        if self.raw.current_time is not None:
            return datetime.utcfromtimestamp(self.raw.current_time)

        return None

    @property
    def commissioning_mode(self) -> int:
        """Return comissioning mode."""
        return self.raw.commissioning_mode

    @property
    def current_time_iso8601(self) -> str:
        """Return current time in iso8601 format."""
        return self.raw.current_time_iso8601

    @property
    def firmware_version(self) -> str:
        """Return gateway firmware version."""
        return self.raw.firmware_version

    @property
    def first_setup(self) -> datetime | None:
        """Return the time when gateway was first set up."""
        if self.raw.first_setup is not None:
            return datetime.utcfromtimestamp(self.raw.first_setup)

        return None

    @property
    def homekit_id(self) -> str:
        """Return homekit id."""
        return self.raw.homekit_id

    @property
    def id(self) -> str:
        """Return the gateway id."""
        return self.raw.id

    @property
    def ntp_server(self) -> str:
        """NTP server in use."""
        return self.raw.ntp_server

    @property
    def ota_type(self) -> int:
        """Return OTA type."""
        return self.raw.ota_type

    @property
    def ota_update_state(self) -> int:
        """Return OTA update state."""
        return self.raw.ota_update_state

    @property
    def pair_status_google_home(self) -> int:
        """Return pairing status Google Home."""
        return self.raw.pair_status_google_home

    @property
    def pair_status_alexa(self) -> int:
        """Return pairing status Amazon Alexa."""
        return self.raw.pair_status_alexa

    @property
    def path(self) -> list[str]:
        """Return path."""
        return [ROOT_GATEWAY, ATTR_GATEWAY_INFO]

    @property
    def time_source(self) -> int:
        """Return time source."""
        return self.raw.time_source

    @property
    def update_progress(self) -> int:
        """Return update status."""
        return self.raw.update_progress

    def set_values(self, values: dict[str, Any]) -> Command[None]:
        """Help set values for mood.

        Returns a Command.
        """
        return Command("put", self.path, values)

    def update(self) -> Command[None]:
        """
        Update the info.

        Returns a Command.
        """

        def process_result(result: TypeRaw) -> None:
            """Define callback to process result."""
            self.raw = GatewayInfoResponse(**result)

        return Command("get", self.path, process_result=process_result)

    def __repr__(self) -> str:
        """Return representation of class object."""
        return "<GatewayInfo>"



"""Group handling."""
from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional

from pydantic import Field

from .color import COLORS
from .command import Command
from .const import (
    ATTR_DEVICE_STATE,
    ATTR_GROUP_ID,
    ATTR_GROUP_MEMBERS,
    ATTR_HS_LINK,
    ATTR_ID,
    ATTR_LIGHT_COLOR_HEX,
    ATTR_LIGHT_COLOR_HUE,
    ATTR_LIGHT_COLOR_SATURATION,
    ATTR_LIGHT_COLOR_X,
    ATTR_LIGHT_COLOR_Y,
    ATTR_LIGHT_DIMMER,
    ATTR_LIGHT_MIREDS,
    ATTR_MOOD,
    ATTR_TRANSITION_TIME,
    RANGE_BRIGHTNESS,
    RANGE_HUE,
    RANGE_MIREDS,
    RANGE_SATURATION,
    RANGE_X,
    RANGE_Y,
    ROOT_GROUPS,
)
from .device import Device
from .error import ColorError
from .mood import Mood
from .resource import ApiResource, ApiResourceResponse, TypeRaw

if TYPE_CHECKING:
    from .gateway import Gateway


class GroupResponse(ApiResourceResponse):
    """Represent API response for a blind."""

    color_hex: Optional[str] = Field(alias=ATTR_LIGHT_COLOR_HEX)
    dimmer: int = Field(alias=ATTR_LIGHT_DIMMER)
    group_members: Dict[str, Dict[str, List[int]]] = Field(alias=ATTR_GROUP_MEMBERS)
    mood_id: str = Field(alias=ATTR_MOOD)
    state: int = Field(alias=ATTR_DEVICE_STATE)


class Group(ApiResource):
    """Represent a group."""

    _model_class: type[GroupResponse] = GroupResponse
    raw: GroupResponse

    def __init__(self, gateway: Gateway, raw: TypeRaw) -> None:
        """Create object of class."""
        super().__init__(raw)
        self._gateway = gateway

    @property
    def path(self) -> list[str]:
        """Path."""
        return [ROOT_GROUPS, str(self.id)]

    @property
    def state(self) -> bool:
        """Boolean representing the light state of the group."""
        return self.raw.state == 1

    @property
    def dimmer(self) -> int:
        """Dimmer value of the group."""
        return self.raw.dimmer

    @property
    def hex_color(self) -> str | None:
        """Return hex color."""
        if self.raw.color_hex:
            return self.raw.color_hex

        return None

    @property
    def member_ids(self) -> list[int]:
        """
        Members of this group.

        A group with devices will look like this:
        {"15002": {"9003": [65536, 65537]}}

        An empty group will look like this:
        {"15002": {"9003": []}}

        If a group is created in the app and no devices are added
        to it, it will immediately be deleted.
        """
        return self.raw.group_members[ATTR_HS_LINK][ATTR_ID]

    @property
    def mood_id(self) -> str:
        """Active mood."""
        return self.raw.mood_id

    def members(self) -> list[Command[Device]]:
        """Return device objects of members of this group."""
        return [self._gateway.get_device(str(dev)) for dev in self.member_ids]

    def add_member(self, memberid: str) -> Command[None]:
        """Add a member to this group."""
        return self._gateway.add_group_member(
            {ATTR_GROUP_ID: self.id, ATTR_ID: [memberid]}
        )

    def remove_member(self, memberid: str) -> Command[None]:
        """Remove a member from this group."""
        return self._gateway.remove_group_member(
            {ATTR_GROUP_ID: self.id, ATTR_ID: [memberid]}
        )

    def moods(self) -> Command[list[Command[Mood]]]:
        """Return mood objects of moods in this group."""
        return self._gateway.get_moods(self.id)

    def mood(self) -> Command[Mood]:
        """Active mood."""
        return self._gateway.get_mood(self.mood_id, mood_parent=self.id)

    def activate_mood(self, mood_id: str) -> Command[None]:
        """Activate a mood."""
        return self.set_values({ATTR_MOOD: mood_id, ATTR_DEVICE_STATE: int(self.state)})

    def set_state(self, state: bool) -> Command[None]:
        """Set state of a group."""
        return self.set_values({ATTR_DEVICE_STATE: int(state)})

    def set_dimmer(
        self, dimmer: int, transition_time: int | None = None
    ) -> Command[None]:
        """Set dimmer value of a group.

        dimmer: Integer between 0..255
        transition_time: Integer representing tenth of a second (default None)
        """
        values = {
            ATTR_LIGHT_DIMMER: dimmer,
        }
        if transition_time is not None:
            values[ATTR_TRANSITION_TIME] = transition_time
        return self.set_values(values)

    def set_color_temp(
        self, color_temp: int, *, index: int = 0, transition_time: int | None = None
    ) -> Command[None]:
        """Set color temp a light."""
        self._value_validate(color_temp, RANGE_MIREDS, "Color temperature")

        values = {ATTR_LIGHT_MIREDS: color_temp}

        if transition_time is not None:
            values[ATTR_TRANSITION_TIME] = transition_time

        return self.set_values(values)

    def set_hex_color(
        self, color: str, transition_time: int | None = None
    ) -> Command[None]:
        """Set hex color of a group."""
        values: dict[str, int | str] = {
            ATTR_LIGHT_COLOR_HEX: color,
        }
        if transition_time is not None:
            values[ATTR_TRANSITION_TIME] = transition_time
        return self.set_values(values)

    def set_hsb(
        self,
        hue: int,
        saturation: int,
        brightness: int | None = None,
        *,
        index: int = 0,
        transition_time: int | None = None,
    ) -> Command[None]:
        """Set HSB color settings of the light."""
        self._value_validate(hue, RANGE_HUE, "Hue")
        self._value_validate(saturation, RANGE_SATURATION, "Saturation")

        values = {ATTR_LIGHT_COLOR_SATURATION: saturation, ATTR_LIGHT_COLOR_HUE: hue}

        if brightness is not None:
            values[ATTR_LIGHT_DIMMER] = brightness
            self._value_validate(brightness, RANGE_BRIGHTNESS, "Brightness")

        if transition_time is not None:
            values[ATTR_TRANSITION_TIME] = transition_time

        return self.set_values(values)

    def set_xy_color(
        self, color_x: int, color_y: int, transition_time: int | None = None
    ) -> Command[None]:
        """Set xy color of a group."""
        self._value_validate(color_x, RANGE_X, "X color")
        self._value_validate(color_y, RANGE_Y, "Y color")

        values = {ATTR_LIGHT_COLOR_X: color_x, ATTR_LIGHT_COLOR_Y: color_y}

        if transition_time is not None:
            values[ATTR_TRANSITION_TIME] = transition_time

        return self.set_values(values)

    def set_predefined_color(
        self, colorname: str, transition_time: int | None = None
    ) -> Command[None]:
        """Set predefined color for group."""
        try:
            color = COLORS[colorname.lower().replace(" ", "_")]
            return self.set_hex_color(color, transition_time=transition_time)
        except KeyError as exc:
            raise ColorError(f"Invalid color specified: {colorname}") from exc

    def _value_validate(
        self, value: int, rnge: tuple[int, int], identifier: str = "Given"
    ) -> None:  # pylint: disable=no-self-use
        """Make sure a value is within a given range."""
        if value is not None and (value < rnge[0] or value > rnge[1]):
            raise ValueError(
                f"{identifier} value must be between {rnge[0]} and {rnge[1]}."
            )

    def __repr__(self) -> str:
        """Return representation of class object."""
        state = "on" if self.state else "off"
        return f"<Group {self.name} - {state}>"



"""COAP implementation using aiocoap."""
from __future__ import annotations

import asyncio
from collections.abc import Callable
from enum import Enum
import json
import logging
from typing import Any, Dict, List, Union, cast, overload

from aiocoap import Context, Message
from aiocoap.credentials import CredentialsMissingError
from aiocoap.error import (
    ConstructionRenderableError,
    Error,
    LibraryShutdown,
    TimeoutError,
)
from aiocoap.numbers.codes import Code
from aiocoap.protocol import BlockwiseRequest

from ..command import Command, T
from ..error import ClientError, RequestTimeout, ServerError
from ..gateway import Gateway

_LOGGER = logging.getLogger(__name__)


class UndefinedType(Enum):
    """Singleton type for use with not set sentinel values."""

    _singleton = 0


_SENTINEL = UndefinedType._singleton  # pylint: disable=protected-access


class APIFactory:
    """ApiFactory."""

    def __init__(
        self,
        host: str,
        psk_id: str = "pytradfri",
        psk: str | None = None,
        internal_create: UndefinedType | None = None,
    ):
        """Create object of class."""
        if internal_create is not _SENTINEL:
            raise ValueError("Use APIFactory.init(…) to initialize APIFactory")

        self._psk = psk
        self._host = host
        self._psk_id = psk_id
        self._observations_err_callbacks: list[Callable[[Exception], None]] = []
        self._protocol: asyncio.Task[Context] | None = None
        self._reset_lock = asyncio.Lock()
        self._shutdown = False

    @classmethod
    async def init(
        cls, host: str, psk_id: str = "pytradfri", psk: str | None = None
    ) -> "APIFactory":
        """Initialize an APIFactory."""
        instance = cls(host, psk_id=psk_id, psk=psk, internal_create=_SENTINEL)
        if psk:
            await instance._update_credentials()
        return instance

    @property
    def psk_id(self) -> str:
        """Return psk id."""
        return self._psk_id

    @property
    def psk(self) -> str | None:
        """Return psk."""
        return self._psk

    async def _get_protocol(self) -> Context:
        """Get the protocol for the request."""
        if self._protocol is None:
            self._protocol = asyncio.create_task(Context.create_client_context())
        return await self._protocol

    async def _reset_protocol(self, exc: BaseException | None = None) -> None:
        """Reset the protocol if an error occurs."""
        skip = self._reset_lock.locked()
        async with self._reset_lock:
            if self._shutdown:
                return
            if skip:
                # The lock was already acquired, so another task was already
                # in the process of resetting the protocol, so we don't need
                # to do it again.
                #
                # This is only here for performance reasons.  It should be
                # safe if the protocol is reset multiple times.
                _LOGGER.debug("Skipping reset: protocol was already being reset")
                return

            _LOGGER.debug("Resetting protocol")

            # Be responsible and clean up.
            protocol = await self._get_protocol()
            await protocol.shutdown()
            self._protocol = None

            # The error callbacks are called when shutting down the protocol.
            # Clear the saved callbacks
            self._observations_err_callbacks.clear()

    async def shutdown(self, exc: Exception | None = None) -> None:
        """Shutdown the API events.

        This should be called before closing the event loop.
        """
        await self._reset_protocol(exc)
        self._shutdown = True

    async def _get_response(
        self, msg: Message, timeout: float | None
    ) -> tuple[BlockwiseRequest, Message]:
        """Perform the request, get the response."""
        try:
            protocol = await self._get_protocol()
            pr_req: BlockwiseRequest = protocol.request(msg)
            pr_resp: Message = await asyncio.wait_for(pr_req.response, timeout)
            return pr_req, pr_resp
        except CredentialsMissingError as exc:
            await self._reset_protocol(exc)
            await self._update_credentials()
            raise ServerError("There was an error with the request.", exc) from exc
        except ConstructionRenderableError as exc:
            raise ClientError("There was an error with the request.", exc) from exc
        except asyncio.TimeoutError as exc:
            await self._reset_protocol(exc)
            await self._update_credentials()
            raise RequestTimeout("Request timed out.", exc) from exc
        except TimeoutError as exc:
            await self._reset_protocol(exc)
            await self._update_credentials()
            raise RequestTimeout("Request timed out.", exc) from exc
        except LibraryShutdown as exc:
            raise ClientError("Protocol is shutdown.", exc) from exc
        except Error as exc:
            await self._reset_protocol(exc)
            await self._update_credentials()
            raise ServerError("There was an error with the request.", exc) from exc
        except asyncio.CancelledError as exc:
            await self._reset_protocol(exc)
            await self._update_credentials()
            raise exc

    async def _execute(self, api_command: Command[T], timeout: float | None) -> T:
        """Execute the command."""
        if api_command.observe:
            await self._observe(api_command, timeout)
            # The observe command result is set by the observe helper method above.
            return api_command.result

        method = api_command.method
        data = api_command.data
        parse_json = api_command.parse_json
        url = api_command.url(self._host)

        kwargs = {}

        if data is not None:
            kwargs["payload"] = json.dumps(data).encode("utf-8")

        api_method = Code.GET
        if method == "put":
            api_method = Code.PUT
        elif method == "post":
            api_method = Code.POST
        elif method == "delete":
            api_method = Code.DELETE
        elif method == "fetch":
            api_method = Code.FETCH
        elif method == "patch":
            api_method = Code.PATCH

        msg = Message(code=api_method, uri=url, **kwargs)

        _LOGGER.debug("Executing %s %s", self._host, api_command)

        _, res = await self._get_response(msg, timeout)
        api_command.process_result(_process_output(res, parse_json))

        return api_command.result

    @overload
    async def request(
        self, api_commands: Command[T], timeout: float | None = None
    ) -> T:
        """Make a request."""
        ...

    @overload
    async def request(
        self, api_commands: list[Command[T]], timeout: float | None = None
    ) -> list[T]:
        """Make a request."""
        ...

    async def request(
        self, api_commands: Command[T] | list[Command[T]], timeout: float | None = None
    ) -> T | list[T]:
        """Make a request."""
        if not isinstance(api_commands, list):
            _LOGGER.debug("REQUEST call single: %s %s", self._host, api_commands)
            result = await self._execute(api_commands, timeout)
            _LOGGER.debug("REQUEST result single: %s", result)

            return result

        _LOGGER.debug("REQUEST call multiple: %s %s", self._host, api_commands)
        commands = (self._execute(api_command, timeout) for api_command in api_commands)
        command_results: list[T] = await asyncio.gather(*commands)
        _LOGGER.debug("REQUEST result multiple: %s", command_results)

        return command_results

    async def _observe(self, api_command: Command[T], timeout: float | None) -> None:
        """Observe an endpoint."""
        duration = api_command.observe_duration
        url = api_command.url(self._host)
        err_callback = api_command.err_callback

        msg = Message(code=Code.GET, uri=url, observe=duration)

        # Note that this is necessary to start observing
        pr_req, pr_rsp = await self._get_response(msg, timeout)

        api_command.process_result(_process_output(pr_rsp))

        def success_callback(res: Message) -> None:
            api_command.process_result(_process_output(res))

        def error_callback(exc: Exception) -> None:
            if isinstance(exc, LibraryShutdown):
                _LOGGER.debug("Protocol is shutdown, stopping observation")
                return

            if err_callback:
                err_callback(exc)

        observation = pr_req.observation
        # The observation is set on the request
        # since we pass a Message with observe set above.
        assert observation is not None
        observation.register_callback(success_callback)
        observation.register_errback(error_callback)
        self._observations_err_callbacks.append(observation.error)

    async def generate_psk(self, security_key: str) -> str:
        """Generate and set a psk from the security key."""
        if not self._psk:
            # Set context once for generating key
            protocol = await self._get_protocol()
            command = Gateway().generate_psk(self._psk_id)
            protocol.client_credentials.load_from_dict(
                {
                    f"coaps://{self._host}:5684/{command.path_str}": {
                        "dtls": {
                            "psk": security_key.encode("utf-8"),
                            "client-identity": "Client_identity".encode("utf-8"),
                        }
                    }
                }
            )

            self._psk = cast(str, await self.request(command))

            # aiocoap has now cached our psk, so it must be reset.
            # We also no longer need the protocol, so this will clean that up.
            await self._reset_protocol()
            await self._update_credentials()

        return self._psk

    async def _update_credentials(self) -> None:
        """Update credentials."""
        if not self._psk:
            # No credentials to reset
            return
        protocol = await self._get_protocol()
        protocol.client_credentials.load_from_dict(
            {
                f"coaps://{self._host}:5684/*": {
                    "dtls": {
                        "psk": self._psk.encode("utf-8"),
                        "client-identity": self._psk_id.encode("utf-8"),
                    }
                }
            }
        )


def _process_output(
    res: Message, parse_json: bool = True
) -> list[Any] | dict[Any, Any] | str | None:
    """Process output."""
    res_payload: str = res.payload.decode("utf-8")
    output = res_payload.strip()

    _LOGGER.debug("Status: %s, Received: %s", res.code, output)

    if not output:
        return None

    # Codes are defined here:
    # https://github.com/chrysn/aiocoap/blob/7b7d43655454682cad6c53e156176d555f3b65a8
    # /aiocoap/numbers/codes.py
    if not res.code.is_successful():
        if 128 <= res.code < 160:
            raise ClientError(f"Gateway payload: {output}. Error code: {res.code}.")
        if 160 <= res.code < 192:
            raise ServerError(f"Gateway payload: {output}. Error code: {res.code}.")

    if not parse_json:
        return output

    return cast(Union[List[Any], Dict[Any, Any]], json.loads(output))



"""Resources for devices."""
from __future__ import annotations

from abc import abstractmethod
from datetime import datetime
from typing import Any, Callable, Dict, List, Optional, Union

from pydantic import BaseModel, Field

from .command import Command
from .const import ATTR_CREATED_AT, ATTR_ID, ATTR_NAME, ATTR_OTA_UPDATE_STATE

# type alias
TypeRaw = Dict[str, Union[str, int, List[Dict[str, Union[str, int]]]]]


class BaseResponse(BaseModel):
    """Represent API base response."""

    id: int = Field(alias=ATTR_ID)


class ApiResourceResponse(BaseResponse):
    """Represent a resource response."""

    name: Optional[str] = Field(alias=ATTR_NAME)
    created_at: Optional[int] = Field(alias=ATTR_CREATED_AT)
    ota_update_state: Optional[int] = Field(alias=ATTR_OTA_UPDATE_STATE)


class ApiResource:
    """Base object for resources returned from the gateway."""

    _model_class: type[ApiResourceResponse] = ApiResourceResponse
    raw: ApiResourceResponse

    def __init__(self, raw: TypeRaw) -> None:
        """Initialize base object."""
        self.raw = self._model_class(**raw)

    @property
    def id(self) -> int:
        """Id."""
        return self.raw.id

    @property
    def name(self) -> str | None:
        """Name."""
        return self.raw.name

    @property
    def created_at(self) -> datetime | None:
        """Return timestamp of creation."""
        created_at = self.raw.created_at

        if created_at is None:
            return None
        return datetime.utcfromtimestamp(created_at)

    @property
    @abstractmethod
    def path(self) -> list[str]:
        """Path to resource."""

    def observe(
        self,
        callback: Callable[[ApiResource], None],
        err_callback: Callable[[Exception], None] | None,
        duration: int = 60,
    ) -> Command[None]:
        """Observe resource and call callback when updated."""

        def observe_callback(value: TypeRaw) -> None:
            """Call when end point is updated.

            Returns a Command.
            """
            self.raw = self._model_class(**value)

            if callback:
                callback(self)

        return Command(
            "get",
            self.path,
            process_result=observe_callback,
            err_callback=err_callback,
            observe=True,
            observe_duration=duration,
        )

    def set_name(self, name: str) -> Command[None]:
        """Set group name."""
        return self.set_values({ATTR_NAME: name})

    def set_values(self, values: Any) -> Command[None]:
        """Help to set values for group.

        Helper to set values for group.
        Returns a Command.
        """
        return Command("put", self.path, values)

    def update(self) -> Command[None]:
        """
        Update the group.

        Returns a Command.
        """

        def process_result(result: TypeRaw) -> None:
            self.raw = self._model_class(**result)

        return Command("get", self.path, process_result=process_result)
